#include "unet.h"

/*
 * Author:
 * Anthony Stahler
 * unfair@unfairnet.com
 *
 * Sorry if this code is a bit messy, I normally don't code in C
 * Please report any bugs to http://bugs.unfairnet.com
 * Thanks
 *
 */

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111-1301  USA
 */

////////////////////////////////////
////	PURPLE FUNCTIONS		////
////////////////////////////////////

static void init_plugin(PurplePlugin *plugin)
{
	if (debugfile == NULL)
	{
		debugfile = fopen("unet_debug.txt","a+");
	}
	purple_debug_info(PLUGIN_ID, "Init UNet\n");

	proto_info.protocol_options = g_list_append(NULL, purple_account_option_string_new("Server", "server", DEFAULT_SERVER));
	proto_info.protocol_options = g_list_append(proto_info.protocol_options, purple_account_option_int_new("Port", "port", DEFAULT_PORT));

	unet_add_chat_commands();
}

static void destroy_plugin(PurplePlugin *plugin)
{
	if (debugfile != NULL)
	{
		fclose(debugfile);
		debugfile = NULL;
	}
	purple_debug_info(PLUGIN_ID, "Destroy UNet\n");
}

static gboolean plugin_load(PurplePlugin *plugin)
{
    return TRUE;
}

static gboolean plugin_unload(PurplePlugin *plugin)
{
	return TRUE;
}

static gchar *buddy_status(PurpleBuddy *pb)
{
	// not actually sure what this is used for
	// it doesn't show up anywhere on the pidgin UI
	// but if you return a pointer or null it crashes
	//
	char *tmpmsg = "";

	if (PURPLE_BUDDY_IS_ONLINE(pb))
	{
		PurplePresence *presence = purple_buddy_get_presence(pb);
		PurpleStatus *status = purple_presence_get_active_status(presence);
		tmpmsg = g_strdup(purple_status_get_attr_string(status, "message"));

		if (tmpmsg == NULL || strlen(tmpmsg) < 1)
		{
			tmpmsg = g_strdup("Online");
		}
	}
	else
	{
		tmpmsg = g_strdup("Offline");
	}

	// purple_util_chrreplace(tmp1, '\n', ' ');

	return tmpmsg;
}

static GList *buddy_status_types(PurpleAccount *account)
{
	GList *types = NULL;

	types = g_list_prepend(types, purple_status_type_new_full(PURPLE_STATUS_OFFLINE, NULL, NULL, FALSE, FALSE, FALSE));
	types = g_list_prepend(types, purple_status_type_new_with_attrs(PURPLE_STATUS_AWAY, NULL, NULL, FALSE, TRUE, FALSE, "message", "Message", purple_value_new(PURPLE_TYPE_STRING), "online", "Online", purple_value_new(PURPLE_TYPE_INT), NULL));
	types = g_list_prepend(types, purple_status_type_new_with_attrs(PURPLE_STATUS_MOBILE, NULL, NULL, FALSE, FALSE, FALSE, "message", "Message", purple_value_new(PURPLE_TYPE_STRING), "online", "Online", purple_value_new(PURPLE_TYPE_INT), NULL));
	types = g_list_prepend(types, purple_status_type_new_with_attrs(PURPLE_STATUS_AVAILABLE, NULL, NULL, FALSE, TRUE, FALSE, "message", "Message", purple_value_new(PURPLE_TYPE_STRING), "online", "Online", purple_value_new(PURPLE_TYPE_INT), NULL));

	return types;
}

static GList *buddy_rclick_menu(PurpleBlistNode *node)
{
	GList *menu = NULL;
	PurpleMenuAction *pma;

	// node types
	// PURPLE_BLIST_GROUP_NODE
	// PURPLE_BLIST_CONTACT_NODE
	// PURPLE_BLIST_BUDDY_NODE
	// PURPLE_BLIST_CHAT_NODE
	// PURPLE_BLIST_OTHER_NODE
	switch (purple_blist_node_get_type(node))
	{
		case PURPLE_BLIST_BUDDY_NODE:
			if (PURPLE_BUDDY_IS_ONLINE((PurpleBuddy*)node))
			{
				// menu action:
				// item
				// callback function
				// userdata passed to the callback
      			// child menu items

				// prepend is faster than append
				//pma = purple_menu_action_new("Profile", PURPLE_CALLBACK(profile_callback), NULL, NULL);
				//menu = g_list_prepend(menu, pma);
			}
			else
			{
				//pma = purple_menu_action_new("Profile2", PURPLE_CALLBACK(profile_callback), NULL, NULL);
				//menu = g_list_prepend(menu, pma);
			}
			break;
		case PURPLE_BLIST_CHAT_NODE:
			break;
		default:
			break;
	}

	return menu;
}

static void user_login(PurpleAccount *acct)
{
	UNetData *ud = g_new0(UNetData, 1);
	ud->salt = NULL;
	ud->rlist = NULL;
	ud->connect_data = NULL;
	ud->in_fd = -1;
	ud->out_fd = -1;
	ud->fd = -1;
	ud->ka_fd = -1;
	ud->salt_fd = -1;
	ud->dc_fd = -1;
	ud->tmprsa = NULL;
	ud->symmetric_enabled = 0;

	havege_init(&ud->rgen);

	ud->current_step = 0;
	PurpleConnection *conn = purple_account_get_connection(acct);

	if(conn != NULL)
	{
		purple_account_clear_current_error(acct);
		//conn->keepalive = 60000;
		conn->proto_data = ud; // add a reference back to our data
		conn->flags |= PURPLE_CONNECTION_HTML;
		//conn->flags |= PURPLE_CONNECTION_FORMATTING_WBFO;
		//conn->flags |= PURPLE_CONNECTION_NO_FONTSIZE;

		ud->sbuff = g_malloc(65536); // max length of packet is 65535
		ud->sbuff_len = 0;
		ud->obuff = g_malloc(65536); // max length of packet is 65535
		ud->obuff_len = 0;

		int i = 0;

		purple_debug_info(PLUGIN_ID,"User login\n");

		// occasionally crashes here in purple_account_get_string if you use ud->acct for the account... odd
		uconnect(acct, purple_account_get_string(acct,"server",DEFAULT_SERVER), purple_account_get_int(acct,"port",DEFAULT_PORT));
	}
	else
	{
		 // massive error, no account/connection
		 // this does seem to happen sometimes, where the PurpleAccount is an invalid reference.  Very odd.
		 close_socket(acct);
	}
}

static void user_logout(PurpleConnection *pc)
{
	if (pc != NULL)
	{
		purple_debug_info(PLUGIN_ID,"User logout\n");
		UNetData *ud = get_ud_from_conn(pc);
		pc->proto_data = NULL;

		if (ud != NULL)
		{
			// account_disconnect destroys conn, no reason to set the state
			//purple_connection_set_state(conn, PURPLE_DISCONNECTED);

			// need to schedule this instead of calling it directly
			// otherwise we risk segfaults if we're in the middle of a callback
			// from a purple function like join chat
			PurpleAccount *acct = purple_connection_get_account(pc);

			if (acct != NULL)
			{
				ud->dc_fd = purple_timeout_add(0, disconnect_cb, acct);
			}
			//purple_account_disconnect(ud->acct);

			close_socket_destroy(ud);
			free_ud(ud); // also sets ud to NULL
		}
	}
	else
	{
		// big problem
	}
}

static int buddy_im(PurpleConnection *pc, const char *who, const char *message, PurpleMessageFlags flags)
{
	// outgoing IM
	int rtnval = -1; // negative return value shows user an error

	PurpleAccount *acct = purple_connection_get_account(pc);

	g_return_val_if_fail(pc != NULL, -1);
	g_return_val_if_fail(acct != NULL, -1);

	if (purple_connection_get_state(pc) == PURPLE_CONNECTED)
	{
		rtnval = send_im(who,message,acct);
	}

	return rtnval;
}

static void buddy_get_info(PurpleConnection *pc, const char *who)
{
	g_return_if_fail(pc != NULL);

	PurpleNotifyUserInfo *tmp_user_info = purple_notify_user_info_new();
	// **** replace this with valid user info later
	// possibly:
	// profile text
	// link to website
	// status
	// online since
	// member since
	// capabilities
	purple_notify_user_info_prepend_pair(tmp_user_info, "Username", who);
	PurpleAccount *acct = purple_connection_get_account(pc);
	PurpleBuddy *pb = purple_find_buddy(acct,who);

	if (PURPLE_BUDDY_IS_ONLINE(pb))
	{
		PurplePresence *presence = purple_buddy_get_presence(pb);
		PurpleStatus *status = purple_presence_get_active_status(presence);

		time_t mtime = purple_status_get_attr_int(status,"online");

		if (mtime > 0)
		{
			char timetrans[30]; // in reality we need about 17, make it 30 just to be safe
			strftime(timetrans,30,"%I:%M%p  %m/%d/%y",localtime(&mtime));
			purple_notify_user_info_add_pair(tmp_user_info, "Online Since", timetrans);
		}

		char *tmpmsg = purple_status_get_attr_string(status, "message");

		if (tmpmsg != NULL)
		{

			if (strlen(tmpmsg) > 0)
			{

				char *tmpstat = purple_markup_strip_html(tmpmsg); // purple_markup_strip_html does g_strdup internally
				purple_notify_user_info_add_pair(tmp_user_info, "Status", tmpstat);
				g_free(tmpstat);
				tmpstat = NULL;
			}
		}
	}



	purple_notify_userinfo(pc,who,tmp_user_info,NULL,NULL);
}

static void buddy_set_status(PurpleAccount *account, PurpleStatus *status)
{
	send_status(account,0);
}

static void buddy_add(PurpleConnection *pc, PurpleBuddy *buddy, PurpleGroup *group)
{
	PurpleAccount *acct = purple_connection_get_account(pc);
	g_return_if_fail(acct != NULL);

	send_packet(acct, construct_packet(acct, 2,15,1,1,construct_subpacket(1,buddy->name,-1,1)));

	// immediately remove the buddy, because of the purple list
	// caching thing - it will be added back later if the server accepts it as okay
	remove_buddy(buddy->name,group->name,acct);
}

static void buddy_remove(PurpleConnection *pc, PurpleBuddy *buddy, PurpleGroup *group)
{
	g_return_if_fail(pc != NULL);
	PurpleAccount *acct = purple_connection_get_account(pc);
	g_return_if_fail(acct != NULL);

	send_packet(acct, construct_packet(acct, 2,15,2,1,construct_subpacket(1,buddy->name,-1,1)));
}

static void buddy_block(PurpleConnection *pc, const char *name)
{
	g_return_if_fail(pc != NULL);
	// ****
}

static void buddy_unblock(PurpleConnection *pc, const char *name)
{
	g_return_if_fail(pc != NULL);
	// ****
}

static void buddy_join_chat(PurpleConnection *pc, GHashTable *components)
{
	g_return_if_fail(pc != NULL);
	PurpleAccount *acct = purple_connection_get_account(pc);
	g_return_if_fail(acct != NULL);

	const char *chatname = g_hash_table_lookup(components, "chatname");

	if (strlen(chatname) > 0)
	{
		send_packet(acct, construct_packet(acct, 2,4,1,1,construct_subpacket(1,chatname,-1,1)));
	}
}

static void buddy_leave_chat(PurpleConnection *pc, int id)
{
	g_return_if_fail(pc != NULL);
	PurpleAccount *acct = purple_connection_get_account(pc);
	g_return_if_fail(acct != NULL);

	PurpleConversation *tmp_conv = purple_find_chat(pc,id);

	send_packet(acct, construct_packet(acct, 2,4,4,1,construct_subpacket(1,tmp_conv->name,-1,1)));
}

static int buddy_chat_im(PurpleConnection *pc, int id, const char *message, PurpleMessageFlags flags)
{
	// outgoing chat message
	int rtnval = -1; // negative return value shows user an error

	PurpleAccount *acct = purple_connection_get_account(pc);

	g_return_val_if_fail(pc != NULL, -1);
	g_return_val_if_fail(acct != NULL, -1);

	// get_state already checks pc for NULL
	if (purple_connection_get_state(pc) == PURPLE_CONNECTED)
	{
		PurpleConversation *tmp_conv = purple_find_chat(pc,id);
		rtnval = send_chat(tmp_conv->name,message,acct);
	}

	return rtnval;
}

static gboolean buddy_offline_support(const PurpleBuddy *buddy)
{
	// Checks whether offline messages to buddy are supported.
	return TRUE;
}

static const char *base_icon(PurpleAccount *account, PurpleBuddy *buddy)
{
	return "unet";
}

static GList *unet_chat_info(PurpleConnection *gc)
{
	GList *m = NULL;
	struct proto_chat_entry *pce;

	pce = g_new0(struct proto_chat_entry, 1); // this is freed by purple
	pce->label = "_Name:";
	pce->identifier = "chatname";
	pce->required = TRUE;
	m = g_list_append(m, pce);
	// **** add optional chat password
	return m;
}

static GHashTable *unet_chat_info_defaults(PurpleConnection *gc, const char *chat_name)
{
	GHashTable *defaults;
	defaults = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, g_free);
	g_hash_table_insert(defaults, "chatname", g_strdup(""));
	return defaults;
}

static const char *buddy_normalize(const PurpleAccount *acct, const char *tmpstr)
{
	if (tmpstr != NULL)
	{
		return unet_normalize(tmpstr);
	}
	else
	{
		return NULL;
	}
}

static void buddy_chat_invite(PurpleConnection *pc, int id, const char *msg, const char *user_name)
{
	PurpleAccount *acct = purple_connection_get_account(pc);
	g_return_if_fail(acct != NULL);
	PurpleConversation *conv = purple_find_chat(pc, id);

	g_return_if_fail(conv != NULL);

	send_packet(acct, construct_packet(acct, 2,4,28,2,construct_subpacket(1,user_name,-1,1),construct_subpacket(2,conv->name,-1,1)));
}

static PurpleRoomlist *fetch_roomlist(PurpleConnection *pc)
{
	PurpleAccount *acct = purple_connection_get_account(pc);
	UNetData *ud = get_ud_from_conn(pc);

	g_return_if_fail(acct != NULL);
	g_return_if_fail(ud != NULL);

	if (ud->rlist != NULL)
	{
		purple_roomlist_set_in_progress(ud->rlist, FALSE);
		purple_roomlist_unref(ud->rlist);
		ud->rlist = NULL;
	}

	ud->rlist = purple_roomlist_new(purple_connection_get_account(pc));
	GList *columns = NULL;
	PurpleRoomlistField *column;

	column = purple_roomlist_field_new(PURPLE_ROOMLIST_FIELD_STRING, "Name", "chatname", TRUE);
	columns = g_list_append(columns, column);

	column = purple_roomlist_field_new(PURPLE_ROOMLIST_FIELD_INT, "Users", "users", FALSE);
	columns = g_list_append(columns, column);

	purple_roomlist_set_fields(ud->rlist, columns);
	purple_roomlist_set_in_progress(ud->rlist, TRUE);
	send_packet(acct, construct_packet(acct, 2,4,49,0));

	return ud->rlist;
}

static void cancel_roomlist(PurpleRoomlist *tmp_rlist)
{
	UNetData *ud = get_ud_from_acct(tmp_rlist->account);
	g_return_if_fail(ud != NULL);

	if (ud->rlist != NULL)
	{
		if (ud->rlist != tmp_rlist)
		{
			purple_roomlist_set_in_progress(ud->rlist, FALSE);
			purple_roomlist_unref(ud->rlist);
		}
	}

	purple_roomlist_set_in_progress(tmp_rlist, FALSE);
	purple_roomlist_unref(tmp_rlist);
	ud->rlist = NULL;
}

////////////////////////////////////
////		FUNCTIONS			////
////////////////////////////////////

static void add_buddy(const char *buddy_name, const char *group_name, PurpleAccount *acct)
{
	// add buddy if they don't exist in the group
	if (purple_find_buddy_in_group(acct, buddy_name, purple_group_new(group_name)) == NULL)
	{
		// can't do it all in one command because of the stupid node saving...
		//purple_blist_add_buddy( purple_buddy_new(plugin_acct, buddy_name, NULL), NULL, purple_group_new(group_name), NULL);

		PurpleGroup *tmppg = purple_find_group(group_name);

		if (tmppg == NULL)
		{
			// group doesn't exist, add it
			tmppg = purple_group_new(group_name);
			purple_blist_add_group(tmppg,NULL);
			purple_blist_node_set_flags(&(tmppg->node), PURPLE_BLIST_NODE_FLAG_NO_SAVE);
		}

		PurpleBuddy *tmppb = purple_buddy_new(acct, buddy_name, NULL);
		purple_blist_node_set_flags(&(tmppb->node), PURPLE_BLIST_NODE_FLAG_NO_SAVE);
		//purple_blist_alias_buddy(tmppb,buddy_name);
		purple_blist_add_buddy( tmppb, NULL, tmppg, NULL);
	}
}

// clear out any buddies that somehow got saved to the purple cache
static void clear_buddy_list(gpointer data, gpointer userdata)
{
	PurpleBuddy *tmpbuddy = (PurpleBuddy*)data;

	if (PURPLE_BLIST_NODE_IS_BUDDY(tmpbuddy))
	{
		purple_blist_remove_buddy(tmpbuddy);
	}
}

static void remove_buddy(const char *buddyname, const char *group_name, PurpleAccount *acct)
{
	// remove the buddy if they exist*
	PurpleBuddy *tmpbuddy = purple_find_buddy_in_group(acct, buddyname, purple_group_new(group_name));

	if (tmpbuddy != NULL)
	{
		purple_blist_remove_buddy(tmpbuddy);
	}

	// remove the group that contained the buddy if it's empty
	PurpleGroup *tmpgroup = purple_find_group(group_name);

	if (tmpgroup != NULL)
	{
		if (tmpgroup->totalsize < 1)
		{
			purple_blist_remove_group(tmpgroup);
		}
	}
}

static UNetData *fetch_node_ud(PurpleBlistNode *node)
{
	UNetData *ud = NULL;
	PurpleBuddy *tmpbuddy = (PurpleBuddy*)node;
	PurpleConnection *tmpconn = purple_account_get_connection(tmpbuddy->account);
	if (tmpconn != NULL)
	{
		ud = tmpconn->proto_data;
	}
	return ud;
}

static void update_connection_step(const char *conn_text, PurpleAccount *acct)
{
	PurpleConnection *conn = purple_account_get_connection(acct);
	UNetData *ud = get_ud_from_conn(conn);

	g_return_if_fail(conn != NULL);
	g_return_if_fail(ud != NULL);

	purple_connection_update_progress(conn, conn_text, ud->current_step, CONNECTION_STEPS);
	ud->current_step++;

	if (ud->current_step >= CONNECTION_STEPS)
	{
		purple_connection_set_state(conn, PURPLE_CONNECTED);
	}

}

static int send_im(const char *touser, const char *message, PurpleAccount *acct)
{
	if (send_packet(acct, construct_packet(acct, 2,3,1,2,construct_subpacket(1,touser,-1,2),construct_subpacket(2,message,-1,2))))
	{
		return 1; // socket wasn't closed
	}
	else
	{
		return -1;
	}
}

static int send_chat(const char *tochat, const char *message, PurpleAccount *acct)
{
	if (send_packet(acct, construct_packet(acct, 2,4,7,2,construct_subpacket(1,tochat,-1,2),construct_subpacket(2,message,-1,2))))
	{
		return 1; // socket wasn't closed
	}
	else
	{
		return -1;
	}
}

static void uconnect(PurpleAccount *acct, const char *server, int port)
{
	PurpleConnection *conn = purple_account_get_connection(acct);
	UNetData *ud = get_ud_from_conn(conn);

	if (conn != NULL && ud != NULL)
	{
		if (server == NULL || strlen(server) < 1)
		{
			server = DEFAULT_SERVER;
		}

		if (port < 1)
		{
			port = DEFAULT_PORT;
		}

		// temporary fix so our old client gets proper disconnect messages
		if (g_ascii_strcasecmp(server, "login.unfairnet.com") == 0)
		{
			server = NEW_DEFAULT_SERVER;
		}

		purple_debug_info(PLUGIN_ID, "Connecting to server:'%s'\n", server);

		update_connection_step("Connecting", acct);

		ud->connect_data = purple_proxy_connect(NULL, acct, server, port, uconnect_cb, acct);

		if (ud->connect_data == NULL)
		{
			// connect error
			purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NETWORK_ERROR, "Could not connect to server.");
			close_socket(acct);
		}
	}
	else
	{
		close_socket(acct);
	}
}

static void uconnect_cb(gpointer userdata, gint fd, const gchar *error_message)
{
	if (userdata != NULL)
	{
		PurpleAccount *acct = (PurpleAccount*)userdata;
		PurpleConnection *conn = purple_account_get_connection(acct);
		UNetData *ud = get_ud_from_conn(conn);

		if (ud != NULL)
		{
			ud->connect_data = NULL;

			if (fd == -1)
			{
				purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NETWORK_ERROR, error_message);
				close_socket(acct);
			}
			else
			{
				ud->fd = fd;
				ud->in_fd = purple_input_add(ud->fd, PURPLE_INPUT_READ, incoming_data, acct);

				// if we're using SSL the server should switch us to it before it sends us the salt
				ud->salt_fd = purple_timeout_add(100, request_salt, acct); 	// if we call this directly instead of scheduling it, our recv socket is
															// closed by the time we return to it for quick errors like max ips
															// and then we get a non-descriptive "Socket Error."
			}
		}
		else
		{
			// if cleanup is run before we get this callback
			// close the connection, we don't need it any more.
			if (fd > -1)
			{
				close(fd);
			}

			// double check that everything else is tidy
			close_socket(acct);
		}
	}
	else
	{
		// major error, why don't we have an account?
		if (fd > -1)
		{
			close(fd);
		}
	}
}

static void request_buddies(PurpleAccount *acct)
{
	update_connection_step("Fetching profile data", acct);

	// request initial buddy list
	send_packet(acct, construct_packet(acct, 2,1,10,0));
}

static gboolean send_packet_finish(gpointer data)
{
	PurpleAccount *acct = (PurpleAccount*)data;
	UNetData *ud = get_ud_from_acct(acct);

	if (ud != NULL)
	{
		ustr *dummystr = g_new0(ustr,1); // freed inside send_packet
		dummystr->len = 0;
		dummystr->data = NULL;

		send_packet(acct, dummystr);
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

static gboolean send_packet(PurpleAccount *acct, ustr *data)
{
	gboolean rtnval = TRUE;

	PurpleConnection *conn = purple_account_get_connection(acct);
	UNetData *ud = get_ud_from_conn(conn);

	if (conn != NULL && ud != NULL)
	{
		if ((data->len + ud->obuff_len) > 65535)
		{
			purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NETWORK_ERROR, "Outgoing buffer full.");
			close_socket(acct);
			rtnval = FALSE;
		}
		else
		{
			if (data->len > 0)
			{
				g_memmove(ud->obuff+ud->obuff_len, data->data, data->len); // copy new packet on to the end of the outgoing buffer
				ud->obuff_len += data->len;
			}

			int bytes_read = 0;
			int bytes_sent = 0;

			while (bytes_sent < ud->obuff_len)
			{
				bytes_read = send(ud->fd, ud->obuff+bytes_sent, ud->obuff_len-bytes_sent, 0);

				if (bytes_read < 1)
				{
					if (errno != EAGAIN && errno != EWOULDBLOCK && errno != EINTR)
					{
						// not one of the three benign errors, close the socket
						purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NETWORK_ERROR, "Socket error.");

						close_socket(acct);
						rtnval = FALSE;
					}
					else
					{
						// shift out buffer
						g_memmove(ud->obuff,ud->obuff+bytes_sent,ud->obuff_len-bytes_sent);
					}
					break;
				}

				bytes_sent += bytes_read;
			}

			ud->obuff_len -= bytes_sent;
		}

		if (ud->obuff_len > 0 && ud->out_fd == -1)
		{
			// schedule resend
			ud->out_fd = purple_timeout_add(1, send_packet_finish, acct);
		}
		else if (ud->obuff_len < 1 && ud->out_fd != -1)
		{
			purple_timeout_remove(ud->out_fd);
			ud->out_fd = -1;
		}
	}
	else
	{
		rtnval = FALSE;
	}

	free_ustr(data);
	data = NULL;

	return rtnval;
}

static void myprintf(char *printdata)
{
	if (debugenabled)
	{
		fprintf(debugfile, printdata, "");
		fflush(debugfile);
	}
}

static void incoming_data(gpointer data, gint source, PurpleInputCondition cond)
{
	PurpleAccount *acct = (PurpleAccount*)data;
	PurpleConnection *conn = purple_account_get_connection(acct);
	UNetData *ud = get_ud_from_conn(conn);

	if(conn != NULL && ud != NULL && acct != NULL)
	{
		int read_size = 1024;
		int buffer_size = read_size + ud->sbuff_len;
		unsigned char *buffer = g_malloc(buffer_size);
		int boffset = 0;

		if (ud->sbuff_len > 0)
		{
			// copy from our old buffer
			g_memmove(buffer+boffset, ud->sbuff, ud->sbuff_len);
		}

		int bytes_read = -1;

		bytes_read = recv(source, buffer+boffset+(buffer_size-read_size),read_size, 0);

		if (bytes_read < 0)
		{
			if (errno != EAGAIN && errno != EWOULDBLOCK && errno != EINTR)
			{
				// not one of the three benign errors, close the socket
				purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NETWORK_ERROR, "Socket error.");
				close_socket(acct);
			}
		}
		else if (bytes_read > 0)
		{
			// we mark sbuff read here in case of errors, so we don't have to copy old data back to it
			ud->sbuff_len = 0;
			buffer_size = (buffer_size-read_size) + bytes_read;

			while (TRUE)
			{
				if (buffer_size >= 2)
				{
					int plen = two_bytes_to_int(buffer+boffset);

					if (buffer_size >= (plen + PACKET_HEAD_LEN))
					{
						// have entire packet in buffer
						if (handle_packet(acct, g_memdup(buffer+boffset, plen + PACKET_HEAD_LEN), plen + PACKET_HEAD_LEN))
						{
							// packet was handled properly, shift our pointer and adjust len
							buffer_size -= (plen + PACKET_HEAD_LEN);

							if (buffer_size > 0)
							{
								// shift pointer
								boffset += (plen + PACKET_HEAD_LEN);
							}
							else
							{
								// no more packets in buffer for now
								break;
							}
						}
						else
						{
							// bad packet, not handled, disconnect user
							purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_OTHER_ERROR, "Invalid packet data.");
							close_socket(acct);
							break;
						}
					}
					else
					{
						// only partial packet, save back to buffer
						g_memmove(ud->sbuff, buffer+boffset, buffer_size);
						ud->sbuff_len = buffer_size;
						break;
					}
				}
				else
				{
					// copy bytes back to buffer
					g_memmove(ud->sbuff, buffer+boffset, buffer_size);
					ud->sbuff_len = buffer_size;
					break;
				}
			}
		}
		else
		{
			// 0 bytes read indicates socket was closed
			// nothing new so dont bother processing again even
			// if there is still data in the buffer
			//purple_account_disconnect(ud->acct);
			//purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NETWORK_ERROR, "Server closed connection.");

			// this should only appear if another error isn't already active
			purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NETWORK_ERROR, "Lost connection.");
			close_socket(acct); // disconnect without error
		}

		g_free(buffer);
		buffer = NULL;
	}
	else
	{
		// we need to be sure this gets closed, otherwise we can get an infinite loop of incoming data
		// if ud gets freed before we closed it, and data is still on the buffer
		// rare, but it can happen
		purple_input_remove(source);
		close(source);
		close_socket(acct);
	}
}

static void close_socket_destroy(UNetData *ud)
{
	if (ud->connect_data != NULL)
	{
		purple_proxy_connect_cancel(ud->connect_data);
		ud->connect_data = NULL;
	}

	if (ud->salt_fd != -1)
	{
		purple_timeout_remove(ud->salt_fd);
		ud->salt_fd = -1;
	}

	if (ud->out_fd != -1)
	{
		purple_timeout_remove(ud->out_fd);
		ud->out_fd = -1;
	}

	if (ud->fd != -1 && ud->fd != -2)
	{
		close(ud->fd);
		ud->fd = -1;
	}

	if (ud->ka_fd != -1)
	{
		purple_timeout_remove(ud->ka_fd);
		ud->ka_fd = -1;
	}

	if (ud->in_fd != -1)
	{
		purple_input_remove(ud->in_fd);
		ud->in_fd = -1;
	}
}

static void close_socket(PurpleAccount *acct)
{
	purple_debug_info(PLUGIN_ID, "Closing socket\n");
	// called by itself, or after a purple_connection_error_reason
	// since that function doesn't immediately stop socket polling and processing
	PurpleConnection *conn = purple_account_get_connection(acct);
	user_logout(conn);
	// this error shouldn't be shown unless we haven't set another already
	if (purple_account_get_current_error(acct) == NULL)
	{
		purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NETWORK_ERROR, "Unknown disconnect.");
	}
}

static gboolean disconnect_cb(gpointer data)
{
	if (data != NULL)
	{
		PurpleAccount *acct = (PurpleAccount*)data;
		purple_account_disconnect(acct);
	}
	return FALSE;
}

static int two_bytes_to_int(const unsigned char *buffer)
{
	return ((buffer[0] * 256) + buffer[1]);
}

static int one_byte_to_int(unsigned char tmpchr)
{
	return (int)tmpchr;
}

static gboolean handle_packet(PurpleAccount *acct, unsigned char *packet, int plen)
{
	gboolean handled = FALSE;
	unsigned char *tmpstr = NULL;
	unsigned char *tmpstr2 = NULL;
	unsigned char *tmpstr3 = NULL;
	int tmpint = 0;

	PurpleConnection *conn = purple_account_get_connection(acct);
	UNetData *ud = get_ud_from_conn(conn);

	if (acct != NULL && conn != NULL && ud != NULL)
	{
		// 2b - packet length
		// 1b - packet type
		// 1b - packet subtype
		//		1b - segment length
		//		xb - segment data
		//		...
		//
		if (plen >= 4)
		{
			int ptype = (int)packet[2];
			int stype = (int)packet[3];

			int offset = 4;

			if (ptype == 1 && stype == 2)
			{
				ud->symmetric_enabled = 1;
			}

			if (ud->symmetric_enabled == 1)
			{
				// dont need to decrypt if there isn't any data
				if (plen > 4)
				{
					// decrypt symmetric, adjust packet and offsets
					unsigned char *tmppkt = packet;
					packet = g_malloc0(plen-offset);

					if (aes_crypt_cbc(&ud->decctx, AES_DECRYPT, plen-offset, ud->decIV, tmppkt+offset, packet) != 0)
					{
						// encryption failed
						PurpleConnection *conn = purple_account_get_connection(acct);
						if (conn != NULL)
						{
							purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Failed to decrypt (s)");
						}
						close_socket(acct);

						g_free(packet);
						packet = NULL;
					}
					else
					{
						plen = plen-offset;
						offset = 0;

						// remove the padding
						plen = pkcs5_len(packet, plen);

						if (plen < 0) // error, bad padding - packet probably wasn't encrypted
						{
							g_free(packet);
							packet = NULL;
						}
					}

					g_free(tmppkt);
					tmppkt = NULL;
				}
			}

			if (packet != NULL)
			{
				switch (ptype)
				{
					case 1:
						switch (stype)
						{
							case 1:
								tmpint = fetch_segment(packet, plen, &(ud->salt), offset, FALSE, 2);
								if (tmpint > 0)
								{
									offset += tmpint;
									ud->tmprsa = g_new0(ustr,1);
									tmpint = fetch_segment(packet, plen, &(ud->tmprsa->data), offset, FALSE, 2);
									ud->tmprsa->len = tmpint-3;

									if (tmpint > 0)
									{
										offset += tmpint;
										rsa_init(&(ud->rsapub), RSA_PKCS_V21, POLARSSL_MD_SHA512);

										tmpint = x509parse_public_key(&(ud->rsapub), ud->tmprsa->data, ud->tmprsa->len);

										if (tmpint == 0)  // len -1 for len byte
										{
											g_slist_foreach(purple_find_buddies(acct,NULL),clear_buddy_list,ud);

											if (purple_find_group("unet") == NULL)
											{
												// we want to add this here so it doesn't get added later and set to NO_SAVE
												// otherwise any chats, etc that we add to the group will be deleted with it
												purple_blist_add_group(purple_group_new("unet"),NULL);
											}

											login(acct->username, acct->password, acct);

										}
										else
										{
											purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Cannot load RSA key.");
											close_socket(acct);
										}

										handled = TRUE;
									}
								}
								else
								{
									// error, no salt..
									purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Salt required.");
									close_socket(acct);
								}
								break;
							case 2:
								{
									tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);

									if (tmpint > 0)
									{
										offset += tmpint;
										char *comphash = enhex(tmpstr, tmpint-2);

										ustr *tmpsalt = g_new0(ustr,1);
										tmpint = fetch_segment(packet, plen, &(tmpsalt->data), offset, TRUE, 1);
										tmpsalt->len = tmpint-2;

										if (tmpint > 0)
										{
											offset += tmpint;


											tmpint = verify_sha512(acct, comphash, tmpsalt);


											if (tmpint == 1)
											{
												tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 1);
												if (tmpint > 0)
												{
													offset += tmpint;
													purple_account_set_alias(acct, (char*)tmpstr2);
													send_status(acct,1);
													request_buddies(acct);
													update_connection_step("Connected", acct);
													// to avoid lag problems we're going to send keepalives twice as often as necessary
													ud->ka_fd = purple_timeout_add_seconds(30, send_keepalive, acct);
												}
												else
												{
													// no formatted name
													purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Name format required.");
													close_socket(acct);
												}
											}
											else
											{
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "RSA verify hash not correct.");
												close_socket(acct);
											}
											handled = TRUE;
										}

										g_free(comphash);
										comphash = NULL;

										free_ustr(tmpsalt);
										tmpsalt = NULL;

										free_ustr(ud->tmprsa);
										ud->tmprsa = NULL;
									}
									else
									{
										purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "RSA verify hash required.");
										close_socket(acct);
									}
								}
								break;
						}
						break;
					case 2:
						switch (stype)
						{
							case 1:
								// disconnect message
								/*
								   • 1 Regular disconnect
								   • 2 Invalid packet
								   • 3 Invalid protocol version
								   • 4 Unregistered username
								   • 5 Invalid password
								   • 6 Banned account
								   • 7 Banned IP
								   • 8 Banned client
								   • 9 Login rate limit.....
								   • 10 Satellite Server ability revoked
								   • 11 Session terminated by administrator
								   • 12 Outdated protocol version, please update before reconnecting
								   • 13 Connection timeout
								   • 14 Database error
								   • 15 All servers full, please try reconnecting in a few minutes
								   • 16 Cookie invalid
								   • 17 Max connections per IP
								   • 18 Invalid hash type
								   • 19 Cannot connect to client
								   • 20 Account not activated
								   • 21 Unexpected error
								   • 22 Invalid encryption type
								   • 23 Username signed on from a different location
								   • 24 Server rebooting
								   • 25 Outdated client version, please update before reconnecting
								*/
								{
									unsigned char *error_code;
									if (fetch_segment(packet, plen, &error_code, offset, FALSE, 1) != -1)
									{
										int tmpcode = (int)error_code[0];

										switch (tmpcode)
										{
											case 1:
												//purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NETWORK_ERROR, "Regular disconnect.");
												break;
											case 2:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_OTHER_ERROR, "Invalid packet.");
												break;
											case 3:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_IMPOSSIBLE, "Outdated protocol version, please update before reconnecting.");
												open_website(DOWNLOAD_URL);
												break;
											case 4:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_INVALID_USERNAME, "Unregistered username.");
												break;
											case 5:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Invalid password.");
												break;
											case 6:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Banned account.");
												break;
											case 7:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_IMPOSSIBLE, "Banned IP.");
												break;
											case 8:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_IMPOSSIBLE, "Banned client.");
												break;
											case 9:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Login rate limit.");
												break;
											case 10:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NETWORK_ERROR, "Satellite Server ability revoked.");
												break;
											case 11:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_OTHER_ERROR, "Session terminated by administrator.");
												break;
											case 12:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NO_SSL_SUPPORT, "SSL connection required.");
												break;
											case 13:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NETWORK_ERROR, "Connection timeout.");
												break;
											case 14:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_OTHER_ERROR, "Database error.");
												break;
											case 15:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "All servers full, please try reconnecting in a few minutes.");
												break;
											case 16:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Cookie invalid.");
												break;
											case 17:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NETWORK_ERROR, "Max connections per IP.");
												break;
											case 18:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Invalid hash type.");
												break;
											case 19:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Cannot connect to client.");
												break;
											case 20:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Account not activated.");
												break;
											case 21:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_OTHER_ERROR, "Unexpected error.");
												break;
											case 22:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_IMPOSSIBLE, "Invalid encryption type.");
												break;
											case 23:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NAME_IN_USE, "Username signed on from a different location.");
												break;
											case 24:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_NETWORK_ERROR, "Server rebooting.");
												break;
											case 25:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_IMPOSSIBLE, "Outdated client version, please update before reconnecting.");
												open_website(DOWNLOAD_URL);
												break;
											case 26:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_OTHER_ERROR, "Rate limit disconnect.");
												break;
											case 27:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_OTHER_ERROR, "Decryption error.");
												break;
											case 28:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_OTHER_ERROR, "Server failed to load encryption key.");
												break;
											default:
												purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_OTHER_ERROR, "Unknown error.");
												break;

										}

										// even after purple_connection_errors the socket keeps running, stop it
										close_socket(acct);
										handled = TRUE;
									}
								}
								break;
						}
						break;
					case 3:
						switch (stype)
						{
							case 1:
								// incoming IM
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 2);
								if (tmpint > 0)
								{
									offset += tmpint;
									tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 2);

									if (tmpint > 0)
									{
										buddy_incoming_im((char*)tmpstr,(char*)tmpstr2,acct,time(NULL));
										handled = TRUE;
									}
								}
								break;
							case 3: // offline incoming IM
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 2);
								if (tmpint > 0)
								{
									offset += tmpint;

									tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 2);

									if (tmpint > 0)
									{
										offset += tmpint;
										tmpint = fetch_segment(packet, plen, &tmpstr3, offset, TRUE, 2);

										if (tmpint > 0)
										{
											time_t tmptime = strtotime((char*)tmpstr3);

											buddy_incoming_im((char*)tmpstr,(char*)tmpstr2,acct,tmptime);
											handled = TRUE;
										}
									}
								}
								break;
							case 5: // system IM
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 2);
								if (tmpint > 0)
								{
									offset += tmpint;

									tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 2);

									if (tmpint > 0)
									{
										offset += tmpint;

										unet_incoming_system_im((char*)tmpstr, (char*)tmpstr2, acct);
										handled = TRUE;
									}
								}
								break;
						}
						break;
					case 4:
						switch (stype)
						{
							case 1: // entered chat
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);
								if (tmpint > 0)
								{
									join_chat((char*)tmpstr,acct);
									handled = TRUE;
								}
								break;
							case 4: // left chat
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);
								if (tmpint > 0)
								{
									leave_chat((char*)tmpstr,acct);
									handled = TRUE;
								}
								break;
							case 7: // chat message
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 2);
								if (tmpint > 0)
								{
									offset += tmpint;

									tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 2);

									if (tmpint > 0)
									{
										offset += tmpint;
										tmpint = fetch_segment(packet, plen, &tmpstr3, offset, TRUE, 2);

										if (tmpint > 0)
										{
											buddy_incoming_chat((char*)tmpstr,(char*)tmpstr2,(char*)tmpstr3,acct,time(NULL));
											handled = TRUE;
										}
									}
								}
								break;
							case 8: // chat log
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 2);
								if (tmpint > 0)
								{
									offset += tmpint;

									tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 2);

									if (tmpint > 0)
									{
										offset += tmpint;
										tmpint = fetch_segment(packet, plen, &tmpstr3, offset, TRUE, 2);

										if (tmpint > 0)
										{

											offset += tmpint;
											char *tmpstr4 = NULL;
											tmpint = fetch_segment(packet, plen, &tmpstr4, offset, TRUE, 2);

											if (tmpint > 0)
											{
												time_t tmptime = strtotime((char*)tmpstr4);

												buddy_incoming_chat((char*)tmpstr,(char*)tmpstr2,(char*)tmpstr3,acct,tmptime);
												handled = TRUE;
											}

											g_free(tmpstr4);
											tmpstr4 = NULL;
										}
									}
								}
								break;
							case 9: // buddy joined chat
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);
								if (tmpint > 0)
								{
									offset += tmpint;

									while (tmpint > 0)
									{
										tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 1);

										if (tmpint > 0)
										{
											offset += tmpint;
											buddy_joined((char*)tmpstr,(char*)tmpstr2,acct);
											handled = TRUE;
										}
									}
								}
								break;
							case 10: // buddy left chat
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);
								if (tmpint > 0)
								{
									offset += tmpint;

									tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 1);

									if (tmpint > 0)
									{
										offset += tmpint;
										buddy_left((char*)tmpstr,(char*)tmpstr2,acct);
										handled = TRUE;
									}
								}
								break;
							case 11:
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);
								if (tmpint > 0)
								{
									offset += tmpint;

									tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 1);

									if (tmpint > 0)
									{
										offset += tmpint;
										tmpint = fetch_segment(packet, plen, &tmpstr3, offset, TRUE, 1);

										if (tmpint > 0)
										{
											buddy_chat_type((char*)tmpstr,(char*)tmpstr2,one_byte_to_int(tmpstr3[0]),acct);
											handled = TRUE;
										}
									}
								}
								break;
							case 17:
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);
								if (tmpint > 0)
								{
									offset += tmpint;
									tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 1);

									if (tmpint > 0)
									{
										set_chat_format((char*)tmpstr,one_byte_to_int(tmpstr2[0]),acct);
										handled = TRUE;
									}
								}
								break;
							case 18:
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);
								if (tmpint > 0)
								{
									offset += tmpint;
									tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 1);

									if (tmpint > 0)
									{
										set_chat_newlines((char*)tmpstr,one_byte_to_int(tmpstr2[0]),acct);
										handled = TRUE;
									}
								}
								break;
							case 20:
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);
								if (tmpint > 0)
								{
									offset += tmpint;

									tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 1);

									if (tmpint > 0)
									{
										unet_incoming_system_chat((char*)tmpstr,(char*)tmpstr2,acct);
										handled = TRUE;
									}
								}

								break;
							case 23:
								// Request box
								// Username has requested permission to join ChatName
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);
								if (tmpint > 0)
								{
									offset += tmpint;

									tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 1);

									if (tmpint > 0)
									{
										show_approval((char*)tmpstr, (char*)tmpstr2, acct);
										handled = TRUE;
									}
								}
								break;
							case 24:
								// Info Message
								// Your request to join ChatName on Username has been Approved/Denied
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);
								if (tmpint > 0)
								{
									offset += tmpint;
									tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 1);

									if (tmpint > 0)
									{
										int tmpjoinint = (int)tmpstr2[0];

										char *result;

										if (tmpjoinint == 0)
										{
											result = "denied";
										}
										else
										{
											result = "approved";
										}
										char *tmpmsg = g_strconcat("Your request to join ", tmpstr, " has been ", result, NULL);

										if (tmpjoinint == 0)
										{
											show_message(purple_account_get_alias(acct),tmpmsg, acct);
										}
										else
										{
											// allow user to join by clicking button on approval box
											show_joinq(purple_account_get_alias(acct),tmpmsg, tmpstr, acct);
										}
										g_free(tmpmsg);
										tmpmsg = NULL;
										handled = TRUE;
									}
								}
								break;
							case 26:
								// chat join failed
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);
								if (tmpint > 0)
								{
									offset += tmpint;
									char *tmperr = g_strconcat("Could not join chat ", tmpstr, NULL);
									show_error(tmperr, "You do not have permission to join this chat", acct);
									g_free(tmperr);
									tmperr = NULL;
									handled = TRUE;
								}
								break;
							case 27:
								// chat join requested
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);
								if (tmpint > 0)
								{
									offset += tmpint;

									char *tmpmsg = g_strconcat("The chat owner must validate your request to join ", tmpstr, NULL);
									show_message(tmpmsg, "You will be notified when your request is approved or denied", acct);
									g_free(tmpmsg);
									tmpmsg = NULL;

									handled = TRUE;
								}
								break;
							case 28:
								// chat invite
								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);
								if (tmpint > 0)
								{
									offset += tmpint;
									tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 1);

									if (tmpint > 0)
									{
										show_invite(tmpstr, tmpstr2, acct);
										handled = TRUE;
									}
								}
								break;
							case 29:
								// join rated
								show_error("Chat join rated", "You are only allowed to join one chat every five seconds", acct);
								handled = TRUE;
								break;
							case 49:
								// chat list
								tmpint = 1;

								while (tmpint > 0)
								{
									tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);

									if (tmpint > 0)
									{
										offset += tmpint;
										add_to_roomlist(ud, (char*)tmpstr);
										handled = TRUE;
									}
								}

								if (!handled)
								{
									// pop up an error, we couldnt find any public rooms...
									show_error("Roomlist fetch failed", "Could not find any public chat rooms", acct);
									handled = TRUE;
								}

								roomlist_complete(ud);
								break;
						}
						break;
					case 15:
						switch (stype)
						{
							case 1: // add buddy
								{
									tmpint = 0;
									while (tmpint >= 0)
									{
										tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);

										if (tmpint > 0)
										{
											add_buddy((char*)tmpstr,"unet",acct);
											offset += tmpint;
										}
									}
									handled = TRUE;
								}
								break;
							case 2: // remove buddy
								{
									tmpint = 0;
									while (tmpint >= 0)
									{
										tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 1);

										if (tmpint > 0)
										{
											remove_buddy((char*)tmpstr,"unet",acct);
											offset += tmpint;
										}
									}
									handled = TRUE;
								}
								break;
							case 5: // status changed
								/*
								Deliver buddy's away status (when changed)
								[15][6][ID][L][l][a][l][b][l][c]

								[a], vb, ASCII buddy name
								[b], 1b, BINARY away type (See 7.8)
								[c], vb, ASCII online since
								[d], vb, ASCII status message
								*/

								tmpint = fetch_segment(packet, plen, &tmpstr, offset, TRUE, 2);

								if (tmpint > 0)
								{
									offset += tmpint;
									tmpint = fetch_segment(packet, plen, &tmpstr2, offset, TRUE, 2);
									if (tmpint > 0)
									{
										int status_type = (int)tmpstr2[0];
										offset += tmpint;

										char *tmpstr4 = NULL;
										tmpint = fetch_segment(packet, plen, &tmpstr4, offset, TRUE, 2);

										if (tmpint > 0)
										{
											time_t tmptime = strtotime((char*)tmpstr4);
											offset += tmpint;
											tmpint = fetch_segment(packet, plen, &tmpstr3, offset, TRUE, 2);

											if (tmpint < 1)
											{
												tmpstr3 = NULL;
											}

											update_buddy_status(status_type,(char*)tmpstr,(char*)tmpstr3,tmptime,acct);
										}

										g_free(tmpstr4);
										tmpstr4 = NULL;
									}

									//purple_notify_user_info_add_pair

							//purple_presence_is_idle(purple_buddy_get_presence(buddy)) ? PURPLE_STATUS_AWAY :
							//	PURPLE_STATUS_AVAILABLE)

									/*
									GSList *tmpbuddies = purple_find_buddies(ud->acct, tmpstr);

									if (tmpbuddies != NULL)
									{
										g_slist_foreach(tmpbuddies,update_buddy_status,NULL);


										// buddy exists on our list


										//PurpleStatus->type->id;



										// get active status
										// create new status
										// set new status active
										// remove old active


										//GList *tmppres = purple_presence_get_statuses(presence);
										//g_list_foreach(tmppres,printstatus,ud);
										//status = purple_presence_get_active_status(presence);
									}
									*/
								}

								handled = TRUE;
								break;
						}
						break;
				}
			}
		}
	}

	g_free(tmpstr);
	tmpstr = NULL;
	g_free(tmpstr2);
	tmpstr2 = NULL;
	g_free(tmpstr3);
	tmpstr3 = NULL;
	g_free(packet);
	packet = NULL;

	return handled;
}

static void show_message(const char *msg, const char *submsg, PurpleAccount *acct)
{
	PurpleConnection *conn = purple_account_get_connection(acct);
	g_return_if_fail(conn != NULL);

	purple_notify_message(conn, PURPLE_NOTIFY_MSG_INFO, "Information", msg, submsg, NULL, NULL);
}

static void show_error(const char *msg, const char *submsg, PurpleAccount *acct)
{
	PurpleConnection *conn = purple_account_get_connection(acct);
	g_return_if_fail(conn != NULL);

	purple_notify_message(conn, PURPLE_NOTIFY_MSG_ERROR, "Error", msg, submsg, NULL, NULL);
}

static void show_joinq(const char *msg, const char *submsg, const char *chat_name, PurpleAccount *acct)
{
	PurpleConnection *conn = purple_account_get_connection(acct);
	g_return_if_fail(conn != NULL);

	ApproveWrapper *tmpwrap = g_new0(ApproveWrapper, 1);
	tmpwrap->acct = acct;
	tmpwrap->chat_name = g_strdup(chat_name);
	tmpwrap->user_name = NULL; // so we don't free an uninit ptr

	purple_request_action(conn, "Information", msg, submsg, PURPLE_DEFAULT_ACTION_NONE, acct, NULL, NULL, tmpwrap, 2, "_Join Now", G_CALLBACK(accept_chat_invite), "_Close", G_CALLBACK(free_aw));
}

static void show_approval(const char *user, const char *chat, PurpleAccount *acct)
{
	PurpleConnection *conn = purple_account_get_connection(acct);
	g_return_if_fail(conn != NULL);

	ApproveWrapper *tmpwrap = g_new0(ApproveWrapper, 1);
	tmpwrap->acct = acct;
	tmpwrap->user_name = g_strdup(user);
	tmpwrap->chat_name = g_strdup(chat);

	char *msg = g_strconcat(user, " has requested permission to join ", chat, NULL);
	char *submsg = "Would you like to approve this request?";

	purple_request_action(conn, "Verify", msg, submsg, PURPLE_DEFAULT_ACTION_NONE, acct, NULL, NULL, tmpwrap, 2, "_Accept", G_CALLBACK(join_approve), "_Deny", G_CALLBACK(join_reject));

	g_free(msg);
	msg = NULL;
}

static void show_invite(const char *from_user, const char *chat_name, PurpleAccount *acct)
{
	PurpleConnection *conn = purple_account_get_connection(acct);
	g_return_if_fail(conn != NULL);

	ApproveWrapper *tmpwrap = g_new0(ApproveWrapper, 1);
	tmpwrap->acct = acct;
	tmpwrap->chat_name = g_strdup(chat_name);
	tmpwrap->user_name = NULL; // so we don't free an uninit ptr

	char *msg = "Chat Invitation";
	char *submsg = g_strconcat(from_user, " has invited you to the chat ", chat_name, NULL);

	purple_request_action(conn, "Chat Invitation", msg, submsg, PURPLE_DEFAULT_ACTION_NONE, acct, NULL, NULL, tmpwrap, 2, "_Accept", G_CALLBACK(accept_chat_invite), "_Cancel", G_CALLBACK(free_aw));

	g_free(submsg);
	submsg = NULL;
}

static void accept_chat_invite(ApproveWrapper *tmpwrap)
{
	g_return_if_fail(tmpwrap != NULL);

	send_packet(tmpwrap->acct, construct_packet(tmpwrap->acct, 2,4,1,1,construct_subpacket(1,tmpwrap->chat_name,-1,1)));

	free_aw(tmpwrap);
	tmpwrap = NULL;
}

static void join_approve(ApproveWrapper *tmpwrap)
{
	g_return_if_fail(tmpwrap != NULL);

	char joinval = (char)1;

	send_packet(tmpwrap->acct, construct_packet(tmpwrap->acct, 2,4,24,3,construct_subpacket(1,tmpwrap->user_name,-1,1),construct_subpacket(2,tmpwrap->chat_name,-1,1),construct_subpacket(3,&joinval,1,1)));

	free_aw(tmpwrap);
	tmpwrap = NULL;
}

static void join_reject(ApproveWrapper *tmpwrap)
{
	g_return_if_fail(tmpwrap != NULL);

	char joinval = (char)0;

	send_packet(tmpwrap->acct, construct_packet(tmpwrap->acct, 2,4,24,3,construct_subpacket(1,tmpwrap->user_name,-1,1),construct_subpacket(2,tmpwrap->chat_name,-1,1),construct_subpacket(3,&joinval,1,1)));

	free_aw(tmpwrap);
	tmpwrap = NULL;
}

static void unet_add_chat_commands()
{
	purple_cmd_register("op", "S", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_CHAT | PURPLE_CMD_FLAG_PRPL_ONLY, PLUGIN_ID, chat_add_op, "op &lt;username&gt;<br />Give a user permission to ban/unban/kick. Must be an owner or admin", NULL);
	purple_cmd_register("unop", "S", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_CHAT | PURPLE_CMD_FLAG_PRPL_ONLY, PLUGIN_ID, chat_remove_op, "unop &lt;username&gt;<br />Remove op permissions. Must be an owner or admin", NULL);
	purple_cmd_register("ban", "S", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_CHAT | PURPLE_CMD_FLAG_PRPL_ONLY, PLUGIN_ID, chat_ban_user, "ban &lt;username&gt;<br />Permanently remove a user from the chat. Must be an op, owner, or admin", NULL);
	purple_cmd_register("remove", "S", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_CHAT | PURPLE_CMD_FLAG_PRPL_ONLY, PLUGIN_ID, chat_ban_user, "remove &lt;username&gt;<br />Permanently remove a user from the chat. Must be an op, owner, or admin", NULL);
	purple_cmd_register("unban", "S", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_CHAT | PURPLE_CMD_FLAG_PRPL_ONLY, PLUGIN_ID, chat_unban_user, "unban &lt;username&gt;<br />Allow a user to join/rejoin the chat. Must be an op, owner, or admin", NULL);
	purple_cmd_register("allow", "S", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_CHAT | PURPLE_CMD_FLAG_PRPL_ONLY, PLUGIN_ID, chat_unban_user, "allow &lt;username&gt;<br />Allow a user to join/rejoin the chat. Must be an op, owner, or admin", NULL);
	purple_cmd_register("kick", "S", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_CHAT | PURPLE_CMD_FLAG_PRPL_ONLY, PLUGIN_ID, chat_kick_user, "kick &lt;username&gt;<br />Temporarily remove a user from the chat. Must be an op, owner, or admin", NULL);
	purple_cmd_register("fontformat", "", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_CHAT | PURPLE_CMD_FLAG_PRPL_ONLY, PLUGIN_ID, chat_font_formatted, "fontformat<br />Toggle to disable or enable text formatting in the chat. Must be an owner or admin", NULL);
	purple_cmd_register("chatmode", "w", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_CHAT | PURPLE_CMD_FLAG_PRPL_ONLY, PLUGIN_ID, chat_mode, "chatmode &lt;public|private|request&gt;<br />Change the chat type. Must be an owner or admin<br />Public - open to all<br />Private - whitelist (invite by owner/op only)<br />Request - a whitelist with join requests sent to the owner", NULL);
	//purple_cmd_register("chatmode", "w", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_CHAT | PURPLE_CMD_FLAG_PRPL_ONLY, PLUGIN_ID, chat_mode, "chatmode &lt;public|private|request|password&gt;<br />Change the chat type. Must be an owner or admin<br />Public - open to all<br />Private - whitelist (invite by owner/op only)<br />Request - a whitelist with join requests sent to the owner<br />Password - password required to join", NULL);
	purple_cmd_register("logging", "", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_CHAT | PURPLE_CMD_FLAG_PRPL_ONLY, PLUGIN_ID, chat_logging, "logging<br />Toggle to disable or enable logging chat messages to the server (default is disabled). Must be an owner or admin", NULL);
	purple_cmd_register("history", "", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_CHAT | PURPLE_CMD_FLAG_PRPL_ONLY, PLUGIN_ID, chat_history, "history<br />Display the last 10 messages from this chat if logging is enabled (default is disabled).", NULL);
	purple_cmd_register("newlines", "", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_CHAT | PURPLE_CMD_FLAG_PRPL_ONLY, PLUGIN_ID, chat_font_newlines, "newlines<br />Toggle to disable or enable newlines in the chat. Must be an owner or admin", NULL);
	purple_cmd_register("invite", "S", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_CHAT | PURPLE_CMD_FLAG_PRPL_ONLY, PLUGIN_ID, chat_invite, "invite &lt;username&gt;<br />Invite a user to join the chat.", NULL);
}

static PurpleCmdRet chat_add_op(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata)
{
	if (!args[0])
	{
		return PURPLE_CMD_RET_FAILED;
	}
	else
	{
		char *tmparg = purple_markup_strip_html(args[0]);

		g_return_val_if_fail(conv->account != NULL, PURPLE_CMD_RET_FAILED);

		send_packet(conv->account, construct_packet(conv->account, 2,4,12,2,construct_subpacket(1,conv->name,-1,1),construct_subpacket(2,tmparg,-1,1)));
		g_free(tmparg);
		tmparg = NULL;
		return PURPLE_CMD_RET_OK;
	}
}

static PurpleCmdRet chat_remove_op(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata)
{
	if (!args[0])
	{
		return PURPLE_CMD_RET_FAILED;
	}
	else
	{
		char *tmparg = purple_markup_strip_html(args[0]);

		g_return_val_if_fail(conv->account != NULL, PURPLE_CMD_RET_FAILED);

		send_packet(conv->account, construct_packet(conv->account, 2,4,13,2,construct_subpacket(1,conv->name,-1,1),construct_subpacket(2,tmparg,-1,1)));
		g_free(tmparg);
		tmparg = NULL;
		return PURPLE_CMD_RET_OK;
	}
}

static PurpleCmdRet chat_ban_user(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata)
{
	if (!args[0])
	{
		return PURPLE_CMD_RET_FAILED;
	}
	else
	{
		char *tmparg = purple_markup_strip_html(args[0]);

		g_return_val_if_fail(conv->account != NULL, PURPLE_CMD_RET_FAILED);

		send_packet(conv->account, construct_packet(conv->account, 2,4,15,2,construct_subpacket(1,conv->name,-1,1),construct_subpacket(2,tmparg,-1,1)));
		g_free(tmparg);
		tmparg = NULL;
		return PURPLE_CMD_RET_OK;
	}
}

static PurpleCmdRet chat_unban_user(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata)
{
	if (!args[0])
	{
		return PURPLE_CMD_RET_FAILED;
	}
	else
	{
		char *tmparg = purple_markup_strip_html(args[0]);

		g_return_val_if_fail(conv->account != NULL, PURPLE_CMD_RET_FAILED);

		send_packet(conv->account, construct_packet(conv->account, 2,4,16,2,construct_subpacket(1,conv->name,-1,1),construct_subpacket(2,tmparg,-1,1)));
		g_free(tmparg);
		tmparg = NULL;
		return PURPLE_CMD_RET_OK;
	}
}

static PurpleCmdRet chat_kick_user(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata)
{
	if (!args[0])
	{
		return PURPLE_CMD_RET_FAILED;
	}
	else
	{
		char *tmparg = purple_markup_strip_html(args[0]);

		g_return_val_if_fail(conv->account != NULL, PURPLE_CMD_RET_FAILED);

		send_packet(conv->account, construct_packet(conv->account, 2,4,14,2,construct_subpacket(1,conv->name,-1,1),construct_subpacket(2,tmparg,-1,1)));
		g_free(tmparg);
		tmparg = NULL;
		return PURPLE_CMD_RET_OK;
	}
}

static PurpleCmdRet chat_font_formatted(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata)
{
	g_return_val_if_fail(conv->account != NULL, PURPLE_CMD_RET_FAILED);

	send_packet(conv->account, construct_packet(conv->account, 2,4,17,1,construct_subpacket(1,conv->name,-1,1)));
	return PURPLE_CMD_RET_OK;
}

static PurpleCmdRet chat_font_newlines(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata)
{
	g_return_val_if_fail(conv->account != NULL, PURPLE_CMD_RET_FAILED);

	send_packet(conv->account, construct_packet(conv->account, 2,4,18,1,construct_subpacket(1,conv->name,-1,1)));
	return PURPLE_CMD_RET_OK;
}

static PurpleCmdRet chat_logging(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata)
{
	g_return_val_if_fail(conv->account != NULL, PURPLE_CMD_RET_FAILED);

	send_packet(conv->account, construct_packet(conv->account, 2,4,50,1,construct_subpacket(1,conv->name,-1,1)));
	return PURPLE_CMD_RET_OK;
}

static PurpleCmdRet chat_history(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata)
{
	g_return_val_if_fail(conv->account != NULL, PURPLE_CMD_RET_FAILED);

	send_packet(conv->account, construct_packet(conv->account, 2,4,51,1,construct_subpacket(1,conv->name,-1,1)));
	return PURPLE_CMD_RET_OK;
}

static PurpleCmdRet chat_mode(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata)
{
	if (!args[0])
	{
		return PURPLE_CMD_RET_FAILED;
	}
	else
	{
		char *tmparg = purple_markup_strip_html(args[0]);
		int cmode = -1;

		if (g_ascii_strncasecmp(tmparg,"pub",3) == 0)
		{
			cmode = 3;
		}
		else if (g_ascii_strncasecmp(tmparg,"priv",4) == 0)
		{
			cmode = 2;
		}
		else if (g_ascii_strncasecmp(tmparg,"req",3) == 0)
		{
			cmode = 1;
		}
		/* **** else if (g_ascii_strncasecmp(tmparg,"pass",4) == 0)
		{
			cmode = 4;
		} */


		if (cmode > 0)
		{
			char modechar = (char)cmode;

			g_return_val_if_fail(conv->account != NULL, PURPLE_CMD_RET_FAILED);

			send_packet(conv->account, construct_packet(conv->account, 2,4,25,2,construct_subpacket(1,conv->name,-1,1),construct_subpacket(2,&modechar,1,1)));
			g_free(tmparg);
			tmparg = NULL;
			return PURPLE_CMD_RET_OK;
		}
		else
		{
			*error = g_strdup_printf("'%s' is an unrecognized chat mode!", tmparg);
			g_free(tmparg);
			tmparg = NULL;
			return PURPLE_CMD_RET_FAILED;
		}
	}
}

static void set_chat_format(char *chatname, int format_type, PurpleAccount *acct)
{
	PurpleConversation *currconv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_CHAT, chatname, acct);

	if (currconv != NULL)
	{
		format_type++;
		// key values are freed for PurpleConversation hash tables, values are not
		g_hash_table_insert(currconv->data, g_strdup("chat_font_format"), GINT_TO_POINTER(format_type));
	}
}

static void set_chat_newlines(char *chatname, int format_type, PurpleAccount *acct)
{
	PurpleConversation *currconv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_CHAT, chatname, acct);

	if (currconv != NULL)
	{
		format_type++;
		// key values are freed for PurpleConversation hash tables, values are not
		g_hash_table_insert(currconv->data, g_strdup("chat_font_newlines"), GINT_TO_POINTER(format_type));
	}
}

static void update_buddy_status(int status_type, char *buddyname, char *status_message, time_t mtime, PurpleAccount *acct)
{
	const char *prplprim = unet_to_prpl_status(status_type);

	if (status_type == UNET_STATUS_OFFLINE)
	{
		purple_prpl_got_user_status(acct,buddyname, prplprim, NULL);
	}
	else
	{
		purple_prpl_got_user_status(acct,buddyname, prplprim, "message", status_message, "online", mtime, NULL);
	}
}

static char *unet_to_prpl_status(int status_type)
{
	char *prplprim;

	switch (status_type)
	{
		case UNET_STATUS_OFFLINE: //offline
			prplprim = purple_primitive_get_id_from_type(PURPLE_STATUS_OFFLINE);
			break;
		case UNET_STATUS_AWAY: // away
			prplprim = purple_primitive_get_id_from_type(PURPLE_STATUS_AWAY);
			break;
		case UNET_STATUS_MOBILE: // mobile
			prplprim = purple_primitive_get_id_from_type(PURPLE_STATUS_MOBILE);
			break;
		case UNET_STATUS_ONLINE: // online
		default:
			prplprim = purple_primitive_get_id_from_type(PURPLE_STATUS_AVAILABLE);
			break;
	}

	return prplprim;
}

static PurpleConvChatBuddyFlags unet_to_prpl_bflags(int buddy_flag)
{
	PurpleConvChatBuddyFlags rtnval = PURPLE_CBFLAGS_NONE;

	switch (buddy_flag)
	{
		case UNET_USER_ADMIN:
			rtnval = PURPLE_CBFLAGS_FOUNDER;
			break;
		case UNET_USER_FOUNDER:
			rtnval = PURPLE_CBFLAGS_OP;
			break;
		case UNET_USER_OP:
			rtnval = PURPLE_CBFLAGS_HALFOP;
			break;
		case UNET_USER_NORMAL:
		default:
			rtnval = PURPLE_CBFLAGS_NONE;
			break;
	}

	return rtnval;
}

static int prpl_to_unet_status(int status_type)
{
	int rtnval = -1;

	switch (status_type)
	{
		case PURPLE_STATUS_OFFLINE:
			rtnval = UNET_STATUS_OFFLINE;
			break;
		case PURPLE_STATUS_AWAY:
			rtnval = UNET_STATUS_AWAY;
			break;
		case PURPLE_STATUS_MOBILE:
			rtnval = UNET_STATUS_MOBILE;
			break;
		case PURPLE_STATUS_AVAILABLE:
		default:
			rtnval = UNET_STATUS_ONLINE;
			break;
	}

	return rtnval;
}

static void send_status(PurpleAccount *acct, int isLogin)
{
	if (acct != NULL)
	{
		PurpleStatus *status = purple_account_get_active_status(acct);
		char *tmpmsg = g_strdup(purple_status_get_attr_string(status, "message"));
		PurpleStatusType *stype = purple_status_get_type(status);
		PurpleStatusPrimitive primitive = purple_status_type_get_primitive(stype);

		int statustype = prpl_to_unet_status((int)primitive);
		char tmpstat = (char)statustype;
		int msglen;

		if (tmpmsg != NULL)
		{
			msglen = strlen(tmpmsg);
		}
		else
		{
			msglen = 0;
		}

		int ptype = 0;
		int subtype = 0;

		if (isLogin == 1)
		{
			ptype = 1;
			subtype = 3;
		}
		else
		{
			ptype = 15;
			subtype = 5;
		}

		send_packet(acct, construct_packet(acct, 2,ptype,subtype,2,construct_subpacket(1,&tmpstat,1,2),construct_subpacket(2,tmpmsg,msglen,2)));

		g_free(tmpmsg);
		tmpmsg = NULL;
	}
}

static ustr *pkcs5_pad(ustr *input)
{
	int blocksize = 16; // 128-bit block cipher
	ustr *rtnval = g_new0(ustr, 1);
	int mval = input->len % blocksize;
	mval = blocksize - mval;

	rtnval->len = input->len + mval;
	rtnval->data = g_malloc(rtnval->len);

	g_memmove(rtnval->data, input->data, input->len);
	int i;
	for (i=input->len;i<rtnval->len;i++)
	{
		rtnval->data[i] = (char)mval;
	}
	free_ustr(input);
	input = NULL;
	return rtnval;
}

static ustr *pkcs5_strip(ustr *input)
{
	int mval = (int)input->data[input->len-1];
	ustr *rtnval = g_new0(ustr, 1);
	rtnval->len = input->len - mval;
	rtnval->data = g_malloc(rtnval->len);

	g_memmove(rtnval->data, input->data, rtnval->len);

	free_ustr(input);
	input = NULL;
	return rtnval;
}

static int pkcs5_len(unsigned char *input, int inputlen)
{
	int mval = (int)input[inputlen-1];
	return (inputlen - mval);
}

static ustr *construct_packet(PurpleAccount *acct, int encrypted, int ptype, int stype, int argc, ...)
{

	UNetData *ud = get_ud_from_acct(acct);
	g_return_if_fail(ud != NULL);

	ustr *rtnarr = g_new0(ustr, 1);
	rtnarr->data = NULL;
	rtnarr->len = 0;

	char pchar = (char)ptype;
	char schar = (char)stype;
	va_list args;

	va_start(args, argc);

	int i;
	for (i=0;i<argc;i++)
	{
		ustr *tmpptr = va_arg(args, ustr*); // grab our next subpacket
		ustr *tmparr = g_new0(ustr, 1); // allocate a new string long enough to hold our previous subpackets plus the new one
		tmparr->len = rtnarr->len + tmpptr->len;
		tmparr->data = g_malloc(tmparr->len);

		g_memmove(tmparr->data, rtnarr->data, rtnarr->len); // copy the old stuff to the new string
		g_memmove((tmparr->data)+(rtnarr->len), tmpptr->data, tmpptr->len); // copy the new segment in

		// free each subpacket as we process it
		free_ustr(tmpptr);
		tmpptr = NULL;

		// free our old stuff
		free_ustr(rtnarr);
		rtnarr = NULL;

		rtnarr = tmparr;
	}

	va_end(args);

	if (rtnarr->len > 0)
	{
		switch (encrypted)
		{
			case 2:
				// symmetric
				rtnarr = pkcs5_pad(rtnarr);

				ustr *tmparr = rtnarr;
				rtnarr = g_new0(ustr, 1);
				rtnarr->len = tmparr->len;
				rtnarr->data = g_malloc0(rtnarr->len);
				if (aes_crypt_cbc(&ud->encctx, AES_ENCRYPT, tmparr->len, ud->encIV, tmparr->data, rtnarr->data) != 0)
				{
					// encryption failed
					PurpleConnection *conn = purple_account_get_connection(acct);
					if (conn != NULL)
					{
						purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Failed to encrypt (s)");
					}
					close_socket(acct);
				}
				free_ustr(tmparr);
				tmparr = NULL;
				break;
			case 1:
			{
				// asymmetric
				ustr *tmparr = rtnarr;
				rtnarr = g_new0(ustr,1);
				rtnarr->len = ud->rsapub.len; // set the out array to the size of the RSA key
				rtnarr->data = g_malloc0(rtnarr->len);

				if (rsa_pkcs1_encrypt(&(ud->rsapub), &havege_rand, &ud->rgen, RSA_PUBLIC, tmparr->len, tmparr->data, rtnarr->data) != 0)
				{
					// encryption failed
					PurpleConnection *conn = purple_account_get_connection(acct);
					if (conn != NULL)
					{
						purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Failed to encrypt (a)");
					}
					close_socket(acct);
				}

				free_ustr(tmparr);
				tmparr = NULL;
			}
			break;
		}
	}

	// add header
	ustr *finptr = g_new0(ustr, 1); // freed inside send_packet
	finptr->len = rtnarr->len + 4;
	finptr->data = g_malloc(finptr->len);

	int_two_bytes(finptr->len-2, finptr->data, 0);
	g_memmove(finptr->data+2, &pchar, 1);
	g_memmove(finptr->data+3, &schar, 1);
	g_memmove(finptr->data+4, rtnarr->data, rtnarr->len);

	free_ustr(rtnarr);
	rtnarr = NULL;

	return finptr;
}

static gboolean request_salt(gpointer data)
{
	PurpleAccount *acct = (PurpleAccount*)data;

	if (acct != NULL)
	{
		int major = 0;
		int minor = 0;
		int revision = 0;

		sscanf(VERSION_MAJOR,"%d",&major);
		sscanf(VERSION_MINOR,"%d",&minor);
		sscanf(VERSION_REV,"%d",&revision);

		char cmajor = (char)major;
		char cminor = (char)minor;
		char crevision = (char)revision;
		//char cauthor = (char)AUTHOR_USER_ID;
		char cprotocol = (char)PROTOCOL_VER;

		char *tmpstr = g_malloc(3);
		int_two_bytes(AUTHOR_USER_ID, tmpstr, 0);

		char *rsa_size = "\x01"; // 3072-bit

		update_connection_step("Authenticating", acct);

		send_packet(acct, construct_packet(acct, 0,1,1,7,construct_subpacket(1,&cprotocol,1,1),
								construct_subpacket(2,UNET_AUTH_NAME,-1,1),
								construct_subpacket(3,&cmajor,1,1),
								construct_subpacket(4,&cminor,1,1),
								construct_subpacket(5,&crevision,1,1),
								construct_subpacket(6,tmpstr,2,1), // author
								construct_subpacket(7,rsa_size,-1,1)
								));

		g_free(tmpstr);
		tmpstr = NULL;
	}

	return FALSE;
}

static ustr *construct_subpacket(int sptype, const char *data, int datalen, int lenbyte)
{
	// datalen is used for strings with nulls
	if (datalen < 0)
	{
		datalen = strlen(data);
	}

	if (lenbyte != 2)
	{
		lenbyte = 1;
	}

	// truncate data to prevent bad packets, just in case.
	// one extra byte off for subpacket types
	if (lenbyte == 1 && datalen > 254)
	{
		datalen = 254;
	}
	else if (lenbyte == 2 && datalen > 65534)
	{
		datalen = 65534;
	}

	ustr *tmpstr = g_new0(ustr, 1); // freed in construct_packet
	tmpstr->len = datalen+1+lenbyte;
	tmpstr->data = g_malloc(tmpstr->len);

	char spchar = (char)sptype;

	if (lenbyte == 2)
	{
		int_two_bytes(tmpstr->len-lenbyte, tmpstr->data, 0);
	}
	else
	{
		char splen = (char)(tmpstr->len-lenbyte);
		g_memmove(tmpstr->data, &splen, 1);
	}
	g_memmove(tmpstr->data+lenbyte, &spchar, 1);
	g_memmove(tmpstr->data+lenbyte+1, data, datalen);

	return tmpstr;
}

static void int_two_bytes(int num, char *rtnval, int pos)
{
	char ch1 = (char)(num / 256);
	char ch2 = (char)(num % 256);
	g_memmove(rtnval, &ch1, 1);
	g_memmove(rtnval+1, &ch2, 1);
}

static char *alt_sha512(PurpleAccount *acct, int hashbytes)
{
	UNetData *ud = get_ud_from_acct(acct);
	g_return_val_if_fail(ud != NULL, g_strdup(""));

	guchar *first = g_malloc(hashbytes);
	guchar *second = g_malloc(hashbytes);

	char *namesalt = "~`!@#$%^&*()_-+=123456x7890";

	// MD5([Username, lowercase no spaces], [Password hash], [Server salt value])

	char *lcname = g_strdup(unet_normalize(acct->username));

	sha4_context ctx;
	sha4_starts(&ctx, USE_SHA512);
	sha4_update(&ctx, (guchar*)acct->password, strlen(acct->password));
	sha4_update(&ctx, (guchar*)namesalt, strlen(namesalt));
	sha4_finish(&ctx, first);

	char *prehash = enhex(first,hashbytes);

	sha4_starts(&ctx, USE_SHA512);
	sha4_update(&ctx, (guchar*)lcname, strlen(lcname));
	sha4_update(&ctx, (guchar*)prehash, strlen(prehash));
	sha4_update(&ctx, (guchar*)ud->salt, 6);
	sha4_finish(&ctx, second);

	g_free(lcname);
	lcname = NULL;
	g_free(prehash);
	prehash = NULL;

	char *rtnval = enhex(second,hashbytes);
	g_free(first);
	first = NULL;
	g_free(second);
	second = NULL;

	return rtnval;
}

static char *unet_hash(PurpleAccount *acct, char *ciphername, int hashbytes)
{
	UNetData *ud = get_ud_from_acct(acct);
	g_return_val_if_fail(ud != NULL, g_strdup(""));

	guchar *first = g_malloc(hashbytes);
	guchar *second = g_malloc(hashbytes);

	char *namesalt = "~`!@#$%^&*()_-+=123456x7890";

	// MD5([Username, lowercase no spaces], [Password hash], [Server salt value])

	char *lcname = g_strdup(unet_normalize(acct->username));

	unsigned int diglen = 0;

	// hash the password first
	PurpleCipherContext *context = purple_cipher_context_new_by_name(ciphername, NULL);
	purple_cipher_context_append(context, (guchar*)acct->password, strlen(acct->password));
	purple_cipher_context_append(context, (guchar*)namesalt, strlen(namesalt));
	purple_cipher_context_digest(context, hashbytes, first, diglen);
	purple_cipher_context_destroy(context);

	char *prehash = enhex(first,hashbytes);

	// create final hash
	context = purple_cipher_context_new_by_name(ciphername, NULL);
	purple_cipher_context_append(context, (guchar*)lcname, strlen(lcname));
	purple_cipher_context_append(context, (guchar*)prehash, strlen(prehash));
	purple_cipher_context_append(context, (guchar*)ud->salt, 6);
	purple_cipher_context_digest(context, hashbytes, second, diglen);
	purple_cipher_context_destroy(context);

	g_free(lcname);
	lcname = NULL;
	g_free(prehash);
	prehash = NULL;

	char *rtnval = enhex(second,hashbytes);

	g_free(first);
	first = NULL;
	g_free(second);
	second = NULL;

	return rtnval;
}

static int verify_sha512(PurpleAccount *acct, char *testhash, ustr *tmpsalt)
{
	int hashbytes = 64;
	UNetData *ud = get_ud_from_acct(acct);
	g_return_val_if_fail(ud != NULL, 0);


	guchar *first = g_malloc(hashbytes);
	guchar *second = g_malloc(hashbytes);

	char *namesalt = "~`!@#$%^&*()_-+=123456x7890";
	char *hashsalt = "92840293!~`#@$^%&*{}_-+q=";

	char *lcname = g_strdup(unet_normalize(acct->username));

	sha4_context ctx;
	sha4_starts(&ctx, USE_SHA512);
	sha4_update(&ctx, (guchar*)acct->password, strlen(acct->password));
	sha4_update(&ctx, (guchar*)namesalt, strlen(namesalt));
	sha4_finish(&ctx, first);

	char *prehash = enhex(first,hashbytes);

	sha4_starts(&ctx, USE_SHA512);
	sha4_update(&ctx, (guchar*)ud->tmprsa->data, ud->tmprsa->len);
	sha4_update(&ctx, (guchar*)tmpsalt->data, tmpsalt->len);
	sha4_update(&ctx, (guchar*)prehash, strlen(prehash));
	sha4_update(&ctx, (guchar*)hashsalt, strlen(hashsalt));
	sha4_finish(&ctx, second);

	g_free(lcname);
	lcname = NULL;
	g_free(prehash);
	prehash = NULL;

	char *rtnval = enhex(second,hashbytes);
	g_free(first);
	first = NULL;
	g_free(second);
	second = NULL;

	if (g_ascii_strcasecmp(rtnval, testhash) == 0)
	{
		g_free(rtnval);
		rtnval = NULL;
		return 1;
	}
	else
	{
		g_free(rtnval);
		rtnval = NULL;
		return 0;
	}
}

static unsigned char *genran(UNetData *ud, int gensize)
{
	unsigned char *rtnval = g_malloc0(gensize);
	int i;

	for (i=0;i<gensize;i++)
	{
		rtnval[i] = (unsigned char)havege_rand(&ud->rgen);
	}

	return rtnval;
}

static void login(const char *user, const char *pass, PurpleAccount *acct)
{
	UNetData *ud = get_ud_from_acct(acct);
	g_return_if_fail(ud != NULL);

	update_connection_step("Verifying", acct);
	char *pw_hash;
	char *proto_type;

	if (purple_ciphers_find_cipher("sha512") != NULL)
	{
		purple_debug_info(PLUGIN_ID,"Login hash: SHA512\n");
		pw_hash = unet_hash(acct,"sha512",64);
		proto_type = "\x03";
	}
	else if (TRUE) // if pidgin doesn't have sha512 just fall back and use our own rather than using a less secure algorithm
	{
		purple_debug_info(PLUGIN_ID,"Login hash: ALT SHA512\n");
		pw_hash = alt_sha512(acct,64);
		proto_type = "\x03";
	}
	else if (purple_ciphers_find_cipher("sha256") != NULL)
	{
		purple_debug_info(PLUGIN_ID,"Login hash: SHA256\n");
		pw_hash = unet_hash(acct,"sha256",32);
		proto_type = "\x02";
	}
	else
	{
		purple_debug_info(PLUGIN_ID,"Login hash: MD5\n");
		pw_hash = unet_hash(acct,"md5",16);
		proto_type = "\x01";
	}

	char *rsa_type = "\x03"; // SHA512
	char *sym_type = "\x01"; // AES

	int tmpransz = 6;
	int tmpkeysz = 32;	// 256 bit key
	int tmpivsz = 16;	// 128 bit IV
	unsigned char *rsa_bytes = genran(ud,tmpransz);
	unsigned char *sym_key = genran(ud,tmpkeysz);
	unsigned char *iv_bytes = genran(ud,tmpivsz);

	if (aes_setkey_enc(&ud->encctx, sym_key, (tmpkeysz*8)) != 0)
	{
		// encryption failed
		PurpleConnection *conn = purple_account_get_connection(acct);
		if (conn != NULL)
		{
			purple_connection_error_reason(conn, PURPLE_CONNECTION_ERROR_AUTHENTICATION_FAILED, "Failed initialize symmetric encryption");
		}
		close_socket(acct);
	}

	aes_setkey_dec(&ud->decctx, sym_key, (tmpkeysz*8));

	int i;
	for (i=0;i<tmpivsz;i++)
	{
		ud->encIV[i] = iv_bytes[i];
		ud->decIV[i] = iv_bytes[i];
	}

	// approximately 133 bytes max packet size, use a 2k RSA key (only workable with a small RSA hash) or greater
	send_packet(acct, construct_packet(acct, 1,1,2,8,construct_subpacket(1,proto_type,-1,1),construct_subpacket(2,pw_hash,-1,1),construct_subpacket(3,user,-1,1),construct_subpacket(4,rsa_type,-1,1),construct_subpacket(5,rsa_bytes,tmpransz,1),construct_subpacket(6,sym_type,-1,1),construct_subpacket(7,sym_key,tmpkeysz,1),construct_subpacket(8,iv_bytes,tmpivsz,1)));

	g_free(pw_hash);
	pw_hash = NULL;
}

static int fetch_segment(const unsigned char *data, int plen, unsigned char **target, int offset, gboolean nullterm, int lenbyte)
{
	int rtnval = -1;

	if (lenbyte != 2)
	{
		lenbyte = 1;
	}

	if (plen >= offset + lenbyte)
	{
		// seglen includes the type byte, but does not include the length byte
		int seglen = 0;

		if (lenbyte == 1)
		{
			seglen = one_byte_to_int(data[offset]);
		}
		else
		{
			seglen = two_bytes_to_int(data + offset);
		}

		if (seglen > 0)
		{
			if (nullterm)
			{
				*target = g_malloc(seglen);
			}
			else
			{
				*target = g_malloc(seglen-1);
			}

			if (plen >= (offset + lenbyte + seglen))
			{
				g_memmove(*target, data+(offset+lenbyte+1), seglen-1);


				if (nullterm == TRUE)
				{
					//target[seglen-1] = '\0';
					g_memmove(*target+(seglen-1), "\0", 1);
					//g_memmove(target+seglen, '\0', 1);
				}
			}
			else
			{
				// **** this should never happen, but we need to put in error handling for it anyway so we don't return a length without data
			}
		}

		rtnval = lenbyte + seglen;
	}

	return rtnval;
}

static gboolean send_keepalive(gpointer data)
{
	PurpleAccount *acct = (PurpleAccount*)data;
	UNetData *ud = get_ud_from_acct(acct);

	if (ud != NULL) // might as well still check for ud, even though we send using acct now
	{
		send_packet(acct, construct_packet(acct, 2,18,2,0));
		return TRUE;
	}
	else
	{
		return FALSE; // stop sending, we can't find the ud
	}
}

static const char *unet_normalize(const char *tmpstr)
{
	static char finbuf[MAX_NORMALIZE+1];
	char *tmpbuf1 = g_ascii_strdown(tmpstr,-1);
	purple_str_strip_char(tmpbuf1, ' '); // this does null term it again

	int cpylen = strlen(tmpbuf1);

	if (cpylen > MAX_NORMALIZE)
	{
		// way too long, nuke it
		cpylen = MAX_NORMALIZE;
	}

	strncpy(finbuf, tmpbuf1, cpylen);
	finbuf[cpylen] = '\0';
	g_free(tmpbuf1);
	tmpbuf1 = NULL;

	return finbuf;
}

static char *enhex(guchar *tmpstr, int length)
{
	char *holder = g_malloc((length*2)+1);
	int offset = 0;
	int i = 0;

	for (i=0;i<length;i++,offset+=2)
	{
		sprintf(holder+offset,"%02x",tmpstr[i]);
	}

	holder[length*2] = '\0'; // zero based, add null to end

	return holder;
}

static void buddy_tooltip(PurpleBuddy *pb, PurpleNotifyUserInfo *user_info, gboolean full)
{
	// **** add support for:
	// logged in time
	// status (available etc)
	// idle time
	// away status
	// username if aliased

	if (PURPLE_BUDDY_IS_ONLINE(pb))
	{
		PurplePresence *presence = purple_buddy_get_presence(pb);
		PurpleStatus *status = purple_presence_get_active_status(presence);

		time_t mtime = purple_status_get_attr_int(status,"online");

		if (mtime > 0)
		{
			// get adjusted time online
			time_t atime = time(NULL) - mtime;

			if (atime < 0)
			{
				atime = 0;
			}

			char *otime = NULL;

			int days = atime / 86400;
			atime = atime % 86400;
			int hours = atime / 3600;
			atime = atime % 3600;
			int minutes = atime / 60;
			atime = atime % 60;
			int seconds = atime;

			if (days > 0)
			{
				otime = g_strdup_printf("%dd%dh%dm%ds", days, hours, minutes, seconds);
			}
			else if (hours > 0)
			{
				otime = g_strdup_printf("%dh%dm%ds", hours, minutes, seconds);
			}
			else if (minutes > 0)
			{
				otime = g_strdup_printf("%dm%ds", minutes, seconds);
			}
			else
			{
				otime = g_strdup_printf("%ds", seconds);
			}

			purple_notify_user_info_add_pair(user_info, "Online Time", otime);
			g_free(otime);
			otime = NULL;
		}

		char *tmpmsg = purple_status_get_attr_string(status, "message");

		if (tmpmsg != NULL)
		{
			if (strlen(tmpmsg) > 0)
			{

				char *tmpstat = purple_markup_strip_html(tmpmsg); // purple_markup_strip_html does g_strdup internally
				purple_notify_user_info_add_pair(user_info, "Status", tmpstat);
				g_free(tmpstat);
				tmpstat = NULL;
			}
		}
	}
}

static void buddy_incoming_im(char *buddyname, char *message, PurpleAccount *acct, time_t mtime)
{
	PurpleConversation *currconv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_IM, buddyname, acct);

	if (currconv == NULL)
	{
		// create new conversation
		currconv = purple_conversation_new(PURPLE_CONV_TYPE_IM, acct, buddyname);
	}

	if (currconv != NULL)
	{
		// exists now
		purple_conv_im_write(currconv->u.im, buddyname, message, PURPLE_MESSAGE_RECV, mtime);
	}
	else
	{
		// error..?
	}
}

static void unet_incoming_system_im(char *buddyname, char *message, PurpleAccount *acct)
{
	PurpleConversation *currconv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_IM, buddyname, acct);

	if (currconv != NULL)
	{
		purple_conv_im_write(currconv->u.im, "system", message, PURPLE_MESSAGE_SYSTEM, time(NULL));
	}
}


static void join_chat(char *chatname, PurpleAccount *acct)
{
	PurpleConnection *conn = purple_account_get_connection(acct);
	g_return_if_fail(conn != NULL);

	serv_got_joined_chat(conn, get_next_chat_id(), chatname);
}

static void leave_chat(char *chatname, PurpleAccount *acct)
{
	PurpleConversation *currconv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_CHAT, chatname, acct);

	if (currconv != NULL)
	{
		purple_conv_chat_left(currconv->u.chat);
	}
}

static void buddy_joined(char *chatname, char *buddyname, PurpleAccount *acct)
{
	PurpleConversation *currconv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_CHAT, chatname, acct);

	if (currconv != NULL)
	{
		purple_conv_chat_add_user(currconv->u.chat, buddyname, NULL, PURPLE_CBFLAGS_NONE, TRUE);
	}
}

static void buddy_chat_type(char *chatname, char *buddyname, int btype, PurpleAccount *acct)
{
	PurpleConversation *currconv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_CHAT, chatname, acct);

	if (currconv != NULL)
	{
		purple_conv_chat_user_set_flags (currconv->u.chat, buddyname, unet_to_prpl_bflags(btype));
 	}
}

static void buddy_left(char *chatname, char *buddyname, PurpleAccount *acct)
{
	PurpleConversation *currconv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_CHAT, chatname, acct);

	if (currconv != NULL)
	{
		purple_conv_chat_remove_user(currconv->u.chat, buddyname, NULL);
	}

}

static void buddy_incoming_chat(char *chatname, char *buddyname, char *message, PurpleAccount *acct, time_t mtime)
{
	PurpleConversation *currconv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_CHAT, chatname, acct);

	if (currconv != NULL)
	{
		PurpleConnection *pconn = purple_conversation_get_gc(currconv);

		if (pconn != NULL)
		{
			// exists now
			gpointer tmp_format = g_hash_table_lookup(currconv->data, "chat_font_format");
			int chat_font_format = 2;
			gpointer tmp_newlines = g_hash_table_lookup(currconv->data, "chat_font_newlines");
			int chat_font_newlines = 2;

			if (tmp_format != NULL)
			{
				chat_font_format = GPOINTER_TO_INT(tmp_format);
			}

			if (tmp_newlines != NULL)
			{
				chat_font_newlines = GPOINTER_TO_INT(tmp_newlines);
			}

			if (chat_font_newlines == 1)
			{
				//purple_str_strip_char(message, '\n'); // this does null term it again
				purple_str_strip_char(message, '\r');
				purple_util_chrreplace(message, '\n', ' '); // we want spaces instead of newlines
				strip_html_string(message, "<br>", 1);
				strip_html_string(message, "<br/>", 1);
			}

			// 0 = 1
			// 1 = 2
			// because c cant tell a NULL from a zero..
			if (chat_font_format == 1)
			{
				char *stripmsg = purple_markup_strip_html(message);
				serv_got_chat_in(pconn, currconv->u.chat->id, buddyname, PURPLE_MESSAGE_RECV, stripmsg, mtime);
				g_free(stripmsg);
				stripmsg = NULL;
			}
			else
			{
				serv_got_chat_in(pconn, currconv->u.chat->id, buddyname, PURPLE_MESSAGE_RECV, message, mtime);
			}
		}
	}
	else
	{
		// error.. getting a message from a chat we aren't in..?
	}
}

static void unet_incoming_system_chat(char *chatname, char *message, PurpleAccount *acct)
{
	PurpleConversation *currconv = purple_find_conversation_with_account(PURPLE_CONV_TYPE_CHAT, chatname, acct);

	if (currconv != NULL)
	{
		system_incoming_chat(message,currconv);
	}
}

static void system_incoming_chat(char *message, PurpleConversation *conv)
{
	purple_conv_chat_write(conv->u.chat, "system", message, PURPLE_MESSAGE_SYSTEM, time(NULL));
}

static time_t strtotime(char *tmpts)
{
	int tmpint = 0;
	time_t rtnval;

	sscanf(tmpts,"%d",&tmpint);

	rtnval = (time_t)tmpint;

	return rtnval;
}

static int get_next_chat_id()
{
	if (chatcounter == 32767)
	{
		chatcounter = 1;
	}
	else
	{
		chatcounter++;
	}

	return chatcounter;
}

static void free_ustr(ustr *tmpstr)
{
	tmpstr->len = 0;
	g_free(tmpstr->data);
	tmpstr->data = NULL;
	g_free(tmpstr);
}

static void free_aw(ApproveWrapper *tmpaw)
{
	tmpaw->acct = NULL;
	g_free(tmpaw->chat_name);
	tmpaw->chat_name = NULL;
	g_free(tmpaw->user_name);
	tmpaw->user_name = NULL;
	g_free(tmpaw);
	tmpaw = NULL;
}

static void free_ud(UNetData *ud)
{
	ud->sbuff_len = 0;
	ud->obuff_len = 0;
	g_free(ud->sbuff);
	ud->sbuff = NULL;
	g_free(ud->obuff);
	ud->obuff = NULL;
	g_free(ud->salt);
	ud->salt = NULL;

	if (ud->rlist != NULL)
	{
		if (purple_roomlist_get_in_progress(ud->rlist))
		{
			purple_roomlist_cancel_get_list(ud->rlist);
		}

		ud->rlist = NULL;
	}

	g_free(ud);
	ud = NULL;
}

static void add_to_roomlist(UNetData *ud, const char *tmpstr)
{
	char *roomname = NULL;
	char *roomppl = NULL;

    roomname = strtok(tmpstr, ":");
    roomppl = strtok(NULL, ":");

    if (roomppl != NULL && ud->rlist != NULL)
    {
		PurpleRoomlistRoom *tmproom = purple_roomlist_room_new(PURPLE_ROOMLIST_ROOMTYPE_ROOM, roomname, NULL);
		purple_roomlist_room_add_field(ud->rlist, tmproom, roomname);
		purple_roomlist_room_add_field(ud->rlist, tmproom, strtol(roomppl, NULL, 10));
		purple_roomlist_room_add(ud->rlist, tmproom);
	}
}

static void roomlist_complete(UNetData *ud)
{
	purple_roomlist_set_in_progress(ud->rlist, FALSE);
}

static UNetData *get_ud_from_conn(PurpleConnection *pc)
{
	UNetData *rtnval = NULL;

	if (PURPLE_CONNECTION_IS_VALID(pc))
	{
		if (pc->proto_data != NULL)
		{
			rtnval = (UNetData*)pc->proto_data;
		}
	}

	return rtnval;
}

static char *strip_html_string(char *text, char *htmlchars, int addspace) // modification of purple_str_strip_char to strip html strings
{
	char *tostrip = g_ascii_strdown(htmlchars,-1);
	purple_str_strip_char(tostrip, ' ');

	int htmllen = strlen(tostrip);

	if (htmllen > 0)
	{
		int htmlpos = 0;
		int startmatch = -1;
		int i, j;

		if (text != NULL)
		{
			for (i = 0, j = 0; text[i]; i++)
			{
				if (tolower(text[i]) == tostrip[htmlpos])
				{
					if (htmlpos == 0)
					{
						startmatch = i;
					}

					htmlpos++;

					if (htmlpos >= htmllen)
					{
						// got full match
						htmlpos=0;
						startmatch = -1;

						if (addspace == 1)
						{
							// add a space where we are stripping out the tag
							text[j++] = ' ';
						}
					}
				}
				else if (htmlpos > 0)
				{
					// inside a match, and this character doesnt match
					if (!g_ascii_isspace(text[i]))
					{
						// not a space, reset and copy all from the start of the match over
						htmlpos = 0;
						int k;
						for (k=startmatch;k<=i;k++)
						{
							text[j++] = text[k];
						}

						startmatch = -1;
					}
				}
				else
				{
					text[j++] = text[i]; //just copy, not in a match
				}
			}

			text[j++] = '\0';
		}
	}

	g_free(tostrip);
	tostrip = NULL;

	return text;
}

static PurpleCmdRet chat_invite(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata)
{
	if (!args[0])
	{
		return PURPLE_CMD_RET_FAILED;
	}
	else
	{
		char *tmparg = purple_markup_strip_html(args[0]);

		g_return_val_if_fail(conv->account != NULL, PURPLE_CMD_RET_FAILED);

		send_packet(conv->account, construct_packet(conv->account, 2,4,28,2,construct_subpacket(1,tmparg,-1,1),construct_subpacket(2,conv->name,-1,1)));

		g_free(tmparg);
		tmparg = NULL;
		return PURPLE_CMD_RET_OK;
	}
}

static void open_website(char *site_url)
{
	PurplePlugin *pp = purple_find_prpl(PLUGIN_ID);
	purple_notify_uri(pp, site_url);
}
/*
static void unet_rsa_test(char *user, char *message)
{
	rsa_context rsa;




	sha4_context ctx;
	sha4_starts(&ctx, USE_SHA512);
	sha4_update(&ctx, (guchar*)acct->password, strlen(acct->password));
	sha4_update(&ctx, (guchar*)namesalt, strlen(namesalt));
	sha4_finish(&ctx, first);
}
*/

static UNetData *get_ud_from_acct(PurpleAccount *acct)
{
	return get_ud_from_conn(purple_account_get_connection(acct));
}
